/* This work was supported by the Director, Office of Science, Office of Advanced Scientific
   Computing Research of the U.S. Department of Energy under contract number DE-AC02-05CH11231.
   Copyright (c) 2014. */

/* hocp_util.hpp
    Utility methods. */

#ifndef HIGHORDERCP_HOCP_UTIL_H
#define HIGHORDERCP_HOCP_UTIL_H

#include <blitz/array.h>
//#include <blitz/tinyvec-et.h> // uncomment if using blitz-0.9

namespace HighOrderCP
{

    template<typename T>
    inline T sqr(T x)
    {
        return x*x;
    }

    // Squared L2 norm of a vector
    template<typename T, int N>
    inline T magsqr(const blitz::TinyVector<T,N>& x)
    {
        T res = x(0)*x(0);
        for (int i = 1; i < N; ++i)
            res += x(i)*x(i);
        return res;
    }

    // L2 norm of a vector
    template<typename T, int N>
    inline T mag(const blitz::TinyVector<T,N>& x)
    {
        T res = x(0)*x(0);
        for (int i = 1; i < N; ++i)
            res += x(i)*x(i);
        return sqrt(res);
    }

    // Returns index of the component of a vector with largest value
    template<typename T, int N>
    int argmax(const blitz::TinyVector<T,N>& x)
    {
        int arg = 0;
        for (int i = 1; i < N; ++i)
            if (x(i) > x(arg))
                arg = i;
        return arg;
    }

    // Template recursion for MultiLoop object
    template<int D>
    inline void incIndices(int* i, const int* min, const int* max, bool& valid)
    {
        if (++i[D-1] >= max[D-1])
        {
            i[D-1] = min[D-1];
            incIndices<D-1>(i, min, max, valid);
        }
    }

    template<>
    inline void incIndices<1>(int* i, const int* min, const int* max, bool& valid)
    {
        if (++i[0] >= max[0])
            valid = false;
    }

    /* MultiLoop implements a nested for-loop with depth known at compile time and can be used
       to loop over an N-dimensional rectangular grid. */
    template<int N>
    class MultiLoop
    {
        const blitz::TinyVector<int,N> min, max;
        blitz::TinyVector<int,N> i;
        bool valid;
    public:
        MultiLoop(const blitz::TinyVector<int,N>& min, const blitz::TinyVector<int,N>& max)
            : min(min), max(max), i(min)
        {
            valid = true;
            for (int dim = 0; dim < N; ++dim)
                valid &= min(dim) <= i(dim) && i(dim) < max(dim);
        }

        inline MultiLoop& operator++()
        {
            incIndices<N>(i.data(), min.data(), max.data(), valid);
            return *this;
        }

        inline const blitz::TinyVector<int,N>& operator()() const
        {
            return i;
        }

        inline operator bool() const
        {
            return valid;
        }

        inline int operator()(int index) const
        {
            return i(index);
        }
    };

} // namespace HighOrderCP

#endif
