/* This work was supported by the Director, Office of Science, Office of Advanced Scientific
   Computing Research of the U.S. Department of Energy under contract number DE-AC02-05CH11231.
   Copyright (c) 2014. */

/* kdtree.hpp
    HighOrderCP::KDTree<T,N> constructs a k-d tree data structure for a given collection of
    points of type blitz::TinyVector<T,N>, where T is typically float or double, and N is 
    the dimension. This particular implementation of the k-d tree has been optimised for the 
    case that the points are situated on a smooth codimension-one surface by using coordinate
    transformations that result in "tight" bounding boxes for some of the nodes in the tree.
    Refer to the paper
        R. I. Saye, High-order methods for computing distances to implicitly defined surfaces,
        Communications in Applied Mathematics and Computational Science, 9(1), 107-141 (2014),
        http://dx.doi.org/10.2140/camcos.2014.9.107
    for more information. */ 

#ifndef HIGHORDERCP_KDTREE_H
#define HIGHORDERCP_KDTREE_H

#include <vector>
#include <algorithm>
#include "hocp_util.hpp"

namespace HighOrderCP
{

    /* A functor for choosing a specific dimension in the std::nth_element selection algorithm */
    template<typename T, int N>
    struct SelectDimension
    {
        const std::vector< blitz::TinyVector<T,N> >& points;
        int dim;
        SelectDimension(const std::vector< blitz::TinyVector<T,N> >& points, int dim) : points(points), dim(dim) {}
        inline bool operator() (int i0, int i1) const
        {
            return points[i0](dim) < points[i1](dim);
        }
    };

    /* Returns the squared-distance from a point x to the hyperrectangle [min,max] (if x is inside the rectangle, the distance is zero) */
    template<typename T, int N>
    T sqrDistBox(const blitz::TinyVector<T,N>& x, const blitz::TinyVector<T,N>& min, const blitz::TinyVector<T,N>& max)
    {
        T dsqr = (T)0;
        for (int i = 0; i < N; ++i)
            if (x(i) < min(i))
                dsqr += sqr(x(i) - min(i));
            else if (x(i) > max(i))
                dsqr += sqr(x(i) - max(i));
        return dsqr;
    }

    template<typename T, int N>
    class KDTree
    {
        std::vector< blitz::TinyVector<T,N> > points;
        std::vector<int> index;
        static const int leaf_size = 16;

        // A Node of the tree is a leaf node iff type=-1. Else if type>=0, then a coordinate transform
        // is applied, otherwise the node is a standard splitting node with two children and no transform
        struct Node
        {
            int type;
            int i0, i1;
            blitz::TinyVector<T,N> xmin, xmax;
        };

        std::vector<Node> nodes;
        std::vector< blitz::TinyVector<blitz::TinyVector<T,N>,N> > transforms;

        // Given a node and a range of points [lb,ub), build the tree
        void build_tree(int nodeIndex, int lb, int ub, bool hasTransformed, int level)
        {
            Node& node = nodes[nodeIndex];

            // Compute bounding box and mean
            blitz::TinyVector<T,N> mean = node.xmin = node.xmax = points[index[lb]];
            for (int i = lb + 1; i < ub; ++i)
            {
                const blitz::TinyVector<T,N>& x = points[index[i]];
                mean += x;
                for (int j = 0; j < N; ++j)
                {
                    if (x(j) < node.xmin(j)) node.xmin(j) = x(j);
                    if (x(j) > node.xmax(j)) node.xmax(j) = x(j);
                }
            }
            mean /= (T)(ub - lb);

            if (ub - lb <= leaf_size)
            {
                // Leaf node
                node.type = -1;
                node.i0 = lb;
                node.i1 = ub;
                return;
            }
        
            // Splitting node: default to splitting along greatest extent
            node.type = -2;
            int axis = argmax<T,N>(node.xmax - node.xmin);

            // Evaluate possibility for coordinate transformation
            if (!hasTransformed && level > 5 && ub - lb >= leaf_size * (1 << 2))
            {
                // Estimate normal
                T holeRadiusSqr = sqr(0.05*max(node.xmax - node.xmin));
                blitz::TinyVector<T,N> n = (T)0.0;
                n(0) = 1.0;
                for (int i = lb; i < ub; ++i)
                {
                    blitz::TinyVector<T,N> x = points[index[i]] - mean;
                    T msqr = magsqr(x);
                    if (msqr > holeRadiusSqr)
                        n -= x * (dot(x,n)/msqr);
                }
                T msqr = magsqr(n);
                if (msqr == 0.0)
                    n(0) = 1.0;
                else
                    n /= sqrt(msqr);

                // Compute alpha range
                T minAlpha = std::numeric_limits<T>::max();
                T maxAlpha = -std::numeric_limits<T>::max();
                for (int i = lb; i < ub; ++i)
                {
                    T alpha = dot(points[index[i]], n);
                    if (alpha > maxAlpha) maxAlpha = alpha;
                    if (alpha < minAlpha) minAlpha = alpha;
                }

                if (maxAlpha - minAlpha < 0.1*max(node.xmax - node.xmin))
                {
                    // Perform transformation: calculate an orthonormal basis using the normal as first axis
                    blitz::TinyVector<blitz::TinyVector<T,N>,N> axes;
                    axes(0) = n;
                    for (int dim = 1, i = 0; dim < N; )
                    {
                        blitz::TinyVector<T,N> y = (T)0.0;
                        y(i) = 1.0;
                        for (int j = 0; j < dim; ++j)
                            y -= dot(y, axes(j))*axes(j);
                        T m = mag(y);
                        if (m > 10.0*std::numeric_limits<T>::epsilon())
                            axes(dim++) = y/m;
                        else
                            ++i;
                        assert(i < N);
                    }

                    // Apply coordinate transformation and calculate new bounding box in order to determine new split direction
                    blitz::TinyVector<T,N> bmin = std::numeric_limits<T>::max();
                    blitz::TinyVector<T,N> bmax = -std::numeric_limits<T>::max();
                    for (int i = lb; i < ub; ++i)
                    {
                        blitz::TinyVector<T,N> x = points[index[i]];
                        for (int dim = 0; dim < N; ++dim)
                        {
                            T alpha = dot(axes(dim), x);
                            points[index[i]](dim) = alpha;
                            if (alpha < bmin(dim)) bmin(dim) = alpha;
                            if (alpha > bmax(dim)) bmax(dim) = alpha;                    
                        }
                    }

                    node.type = transforms.size();
                    transforms.push_back(axes);
                    axis = argmax<T,N>(bmax - bmin);
                    hasTransformed = true;
                }
            }

            // Use median as the split
            int m = (lb + ub)/2;
            std::nth_element(index.begin() + lb, index.begin() + m, index.begin() + ub, SelectDimension<T,N>(points, axis));

            int i0 = node.i0 = nodes.size();
            int i1 = node.i1 = nodes.size() + 1;
            nodes.push_back(Node());
            nodes.push_back(Node());
            build_tree(i0, lb, m, hasTransformed, level + 1);
            build_tree(i1, m, ub, hasTransformed, level + 1);
        }

        struct ClosestPoint
        {
            blitz::TinyVector<T,N> x;
            T distsqr;
            int ind;
        };

        // Recursive function for searching the tree for the closest point to a given point cp.x
        void search(const Node& node, ClosestPoint& cp) const
        {
            if (node.type == -1)
            {
                // Leaf node
                for (int j = node.i0; j < node.i1; ++j)
                {
                    T dsqr = magsqr<T,N>(points[j] - cp.x);
                    if (dsqr < cp.distsqr)
                    {
                        cp.distsqr = dsqr;
                        cp.ind = j;
                    }
                }
            }
            else
            {
                // Non-leaf node
                if (node.type >= 0)
                {
                    // Transform query point to new coordinate system
                    const blitz::TinyVector<blitz::TinyVector<T,N>,N>& axes = transforms[node.type];
                    blitz::TinyVector<T,N> x = cp.x;
                    for (int dim = 0; dim < N; ++dim)
                        cp.x(dim) = dot(axes(dim), x);
                }

                T dleft = sqrDistBox(cp.x, nodes[node.i0].xmin, nodes[node.i0].xmax);
                T dright = sqrDistBox(cp.x, nodes[node.i1].xmin, nodes[node.i1].xmax);
                if (dleft < dright)
                {
                    if (dleft < cp.distsqr)
                    {
                        search(nodes[node.i0], cp);
                        if (dright < cp.distsqr)
                            search(nodes[node.i1], cp);
                    }
                }
                else
                {
                    if (dright < cp.distsqr)
                    {
                        search(nodes[node.i1], cp);
                        if (dleft < cp.distsqr)
                            search(nodes[node.i0], cp);
                    }
                }

                if (node.type >= 0)
                {
                    // Transform query point back to old coordinate system. This is about 5% faster than storing
                    // the old value of cp.x on the stack
                    const blitz::TinyVector<blitz::TinyVector<T,N>,N>& axes = transforms[node.type];
                    blitz::TinyVector<T,N> x = cp.x;
                    cp.x = axes(0)*x(0);
                    for (int dim = 1; dim < N; ++dim)
                        cp.x += axes(dim)*x(dim);
                }
            }
        }

    public:

        // Construct a KDTree from a given collection of points, p (the given points are not overwritten)
        KDTree(const std::vector< blitz::TinyVector<T,N> >& p)
        {
            size_t len = p.size();
            if (len == 0)
                return;
            points = p;
            index.resize(len);
            for (int i = 0; i < (int)len; ++i)
                index[i] = i;
            nodes.push_back(Node());
            build_tree(0, 0, len, false, 0);
            // Rearrange so that points in leaf nodes are contiguous in memory. Based on tree
            // construction, both nodes & point ranges will be organised depth-first.
            std::vector< blitz::TinyVector<T,N> > pointscopy(points);
            for (size_t i = 0; i < len; ++i)
                points[i] = pointscopy[index[i]];
        }

        // Search the tree for the closest point to x, where the search is restricted to all points that are
        // within a squared distance of rsqr to x. Returns -1 if no such point exists, otherwise the index
        // of the closest point in the original array p passed to the KDTree constructor is returned.
        int nearest(const blitz::TinyVector<T,N>& x, T rsqr = std::numeric_limits<T>::max()) const
        {
            if (nodes.empty())
                return -1;
            ClosestPoint cp;
            cp.x = x;
            cp.distsqr = rsqr;
            cp.ind = -1;
            search(nodes[0], cp);
            if (cp.ind >= 0)
                return index[cp.ind];
            else
                return -1;
        }
    };

} // namespace HighOrderCP

#endif
