/* This work was supported by the Director, Office of Science, Office of Advanced Scientific
   Computing Research of the U.S. Department of Energy under contract number DE-AC02-05CH11231.
   Copyright (c) 2014. */

/* poly.hpp
    Implements the least-squares polynomial interpolation algorithms for 10 different
    classes of polynomial in 2D and 3D. */

#ifndef HIGHORDERCP_POLY_H
#define HIGHORDERCP_POLY_H

#include "hocp_util.hpp"

namespace HighOrderCP
{

    // Calculate the coefficients for a polynomial in N dimensions, based on a given stencil of O points
    // evaluated on a rectangular grid, where M is the number of coefficients in the polynomial
    template<int N, int M, int O, typename F>
    void calculateCoefficients(blitz::TinyVector<double,M>& c, const blitz::TinyVector<int,N>* stencil, const double* matrix, const blitz::TinyVector<int,N>& gp, const F& phi)
    {
        blitz::TinyVector<double,O> values;
        for (int i = 0; i < O; ++i)
            values(i) = phi(blitz::TinyVector<int,N>(gp + stencil[i]));
        for (int i = 0; i < M; ++i)
        {
            c(i) = 0.0;
            for (int j = 0; j < O; ++j)
                c(i) += matrix[i*O + j]*values(j);
        }
    }

    // Two-dimensional polynomials have their own namespace
    namespace Poly2D
    {
        // The data for these are defined in poly.cpp
        extern const blitz::TinyVector<int,2> stencil12points[12];
        extern const blitz::TinyVector<int,2> stencil24points[24];
        extern const double polyDegree2Inverse[6*12];
        extern const double polyDegree3Inverse[10*12];
        extern const double polyDegree4Inverse[15*24];
        extern const double polyDegree5Inverse[21*24];
        extern const double bicubicInverse[16*16];

        /* Two-dimensional, degree 2 polynomial of the form
             p(x,y) = c0 + c1 x + c2 y + c3 x^2 + c4 x y + c5 y^2
           based on a 12-point stencil. Pseudoinverse has rank 6. */
        struct PolyDegree2
        {
            enum { order = 3 };

            blitz::TinyVector<double,6> c;

            PolyDegree2() {}

            template<typename F>
            PolyDegree2(const blitz::TinyVector<int,2>& i, const F& phi, const blitz::TinyVector<double,2>& dx)
            {
                calculateCoefficients<2,6,12>(c, stencil12points, polyDegree2Inverse, i, phi);
                blitz::TinyVector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
            }

            double operator() (const blitz::TinyVector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*c(5)) + x(0)*(c(1) + x(1)*c(4) + x(0)*c(3));
            }

            blitz::TinyVector<double,2> grad(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,2>(c(1) + x(1)*c(4) + x(0)*(2.0*c(3)), c(2) + x(1)*(2.0*c(5)) + x(0)*c(4));
            }

            blitz::TinyVector<double,3> hessian(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,3>(2.0*c(3), c(4), 2.0*c(5));
            }
        };

        /* Two-dimensional, degree 3 polynomial of the form
             p(x,y) = c0 + c1 x + c2 y + c3 x^2 + c4 x y + c5 y^2 + c6 x^3 + c7 x^2 y + c8 x y^2 + c9 y^3
           based on a 12-point stencil. Pseudoinverse has rank 10.*/
        struct PolyDegree3
        {
            enum { order = 4 };

            blitz::TinyVector<double,10> c;

            PolyDegree3() {}

            template<typename F>
            PolyDegree3(const blitz::TinyVector<int,2>& i, const F& phi, const blitz::TinyVector<double,2>& dx)
            {
                calculateCoefficients<2,10,12>(c, stencil12points, polyDegree3Inverse, i, phi);
                blitz::TinyVector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(7) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(8) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(9) *= dxinv(1)*dxinv(1)*dxinv(1);
            }

            double operator() (const blitz::TinyVector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*(c(5) + x(1)*c(9))) + x(0)*(c(1) + x(1)*(c(4) + x(1)*c(8)) + x(0)*(c(3) + x(1)*c(7) + x(0)*c(6)));
            }

            blitz::TinyVector<double,2> grad(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,2>(c(1) + x(1)*(c(4) + x(1)*c(8)) + x(0)*(2.0*c(3) + x(1)*(2.0*c(7)) + x(0)*(3.0*c(6))),
                    c(2) + x(1)*(2.0*c(5) + x(1)*(3.0*c(9))) + x(0)*(c(4) + x(1)*(2.0*c(8)) + x(0)*c(7)));
            }

            blitz::TinyVector<double,3> hessian(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,3>(2.0*c(3) + x(1)*(2.0*c(7)) + x(0)*(6.0*c(6)), c(4) + x(1)*(2.0*c(8)) + x(0)*(2.0*c(7)), 2.0*c(5) + x(1)*(6.0*c(9)) + x(0)*(2.0*c(8)));
            }
        };

        /* Two-dimensional, degree 4 polynomial of the form 
             p(x,y) = c0 + c1 x + c2 y + c3 x^2 + c4 x y + c5 y^2 + c6 x^3 + c7 x^2 y + c8 x y^2 + c9 y^3 +
                      c10 x^4 + c11 x^3 y + c12 x^2 y^2 + c13 x y^3 + c14 y^4
           based on a 24-point stencil. Pseudoinverse has rank 15. (Note that for a stencil based on a 4 by 4
           patch, as used in bicubic, the Vandermonde matrix has rank 13.) */
        struct PolyDegree4
        {
            enum { order = 5 };

            blitz::TinyVector<double,15> c;

            PolyDegree4() {};

            template<typename F>
            PolyDegree4(const blitz::TinyVector<int,2>& i, const F& phi, const blitz::TinyVector<double,2>& dx)
            {
                calculateCoefficients<2,15,24>(c, stencil24points, polyDegree4Inverse, i, phi);
                blitz::TinyVector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(7) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(8) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(9) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
            }

            double operator() (const blitz::TinyVector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*(c(5) + x(1)*(c(9) + x(1)*c(14)))) + x(0)*(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*c(13))) + x(0)*(c(3) + x(1)*(c(7) + x(1)*c(12)) + x(0)*(c(6) + x(1)*c(11) + x(0)*c(10))));
            }

            blitz::TinyVector<double,2> grad(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,2>(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*c(13))) + x(0)*(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12))) + x(0)*(3.0*c(6) + x(1)*(3.0*c(11)) + x(0)*(4.0*c(10)))),
                    c(2) + x(1)*(2.0*c(5) + x(1)*(3.0*c(9) + x(1)*(4.0*c(14)))) + x(0)*(c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13))) + x(0)*(c(7) + x(1)*(2.0*c(12)) + x(0)*c(11))));
            }

            blitz::TinyVector<double,3> hessian(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,3>(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12))) + x(0)*(6.0*c(6) + x(1)*(6.0*c(11)) + x(0)*(12.0*c(10))),
                    c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13))) + x(0)*(2.0*c(7) + x(1)*(4.0*c(12)) + x(0)*(3.0*c(11))),
                    2.0*c(5) + x(1)*(6.0*c(9) + x(1)*(12.0*c(14))) + x(0)*(2.0*c(8) + x(1)*(6.0*c(13)) + x(0)*(2.0*c(12))));
            }
        };

        /* Two-dimensional, degree 5 polynomial of the form 
             p(x,y) = c0 + c1 x + c2 y + ... + c20 y^5
           based on a 24-point stencil. Pseudoinverse has rank 21.*/
        struct PolyDegree5
        {
            enum { order = 6 };

            blitz::TinyVector<double,21> c;

            PolyDegree5() {};

            template<typename F>
            PolyDegree5(const blitz::TinyVector<int,2>& i, const F& phi, const blitz::TinyVector<double,2>& dx)
            {
                calculateCoefficients<2,21,24>(c, stencil24points, polyDegree5Inverse, i, phi);
                blitz::TinyVector<double,2> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(0)*dxinv(0);
                c(4) *= dxinv(0)*dxinv(1);
                c(5) *= dxinv(1)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(7) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(8) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(9) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(15) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(16) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(17) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(18) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(19) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(20) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
            }

            double operator() (const blitz::TinyVector<double,2>& x) const
            {
                return c(0) + x(1)*(c(2) + x(1)*(c(5) + x(1)*(c(9) + x(1)*(c(14) + x(1)*c(20))))) + x(0)*(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*(c(13) + x(1)*c(19)))) + x(0)*(c(3) + x(1)*(c(7) + x(1)*(c(12) + x(1)*c(18))) + x(0)*(c(6) + x(1)*(c(11) + x(1)*c(17)) + x(0)*(c(10) + x(1)*c(16) + x(0)*c(15)))));
            }

            blitz::TinyVector<double,2> grad(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,2>(c(1) + x(1)*(c(4) + x(1)*(c(8) + x(1)*(c(13) + x(1)*c(19)))) + x(0)*(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12) + x(1)*(2.0*c(18)))) + x(0)*(3.0*c(6) + x(1)*(3.0*c(11) + x(1)*(3.0*c(17))) + x(0)*(4.0*c(10) + x(1)*(4.0*c(16)) + x(0)*(5.0*c(15))))),
                    c(2) + x(1)*(2.0*c(5) + x(1)*(3.0*c(9) + x(1)*(4.0*c(14) + x(1)*(5.0*c(20))))) + x(0)*(c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13) + x(1)*(4.0*c(19)))) + x(0)*(c(7) + x(1)*(2.0*c(12) + x(1)*(3.0*c(18))) + x(0)*(c(11) + x(1)*(2.0*c(17)) + x(0)*c(16)))));
            }

            blitz::TinyVector<double,3> hessian(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,3>(2.0*c(3) + x(1)*(2.0*c(7) + x(1)*(2.0*c(12) + x(1)*(2.0*c(18)))) + x(0)*(6.0*c(6) + x(1)*(6.0*c(11) + x(1)*(6.0*c(17))) + x(0)*(12.0*c(10) + x(1)*(12.0*c(16)) + x(0)*(20.0*c(15)))),
                    c(4) + x(1)*(2.0*c(8) + x(1)*(3.0*c(13) + x(1)*(4.0*c(19)))) + x(0)*(2.0*c(7) + x(1)*(4.0*c(12) + x(1)*(6.0*c(18))) + x(0)*(3.0*c(11) + x(1)*(6.0*c(17)) + x(0)*(4.0*c(16)))),
                    2.0*c(5) + x(1)*(6.0*c(9) + x(1)*(12.0*c(14) + x(1)*(20.0*c(20)))) + x(0)*(2.0*c(8) + x(1)*(6.0*c(13) + x(1)*(12.0*c(19))) + x(0)*(2.0*c(12) + x(1)*(6.0*c(18)) + x(0)*(2.0*c(17)))));
            }
        };

        /* Two-dimensional, bicubic interpolant guaranteeing C^1 continuity between cells on a rectangular
           grid, using second order finite differences evaluated at gird points to ensure continuity of
           the gradient across grid cells. This interpolant is equivalent to that discussed in D. L. Chopp, 
           Some improvements of the Fast Marching Method, SIAM Journal Scientific Computing 23(1) (2001) 230244. */
        struct Bicubic
        {
            enum { order = 3 };

            blitz::TinyVector<double,16> c;

            template<typename F>
            Bicubic(const blitz::TinyVector<int,2>& i, const F& phi, const blitz::TinyVector<double,2>& dx)
            {
                blitz::TinyVector<int,2> stencilext = 4;
                blitz::Array<double,2> stencil(stencilext);

                for (MultiLoop<2> b(0, 4); b; ++b)
                {
                    blitz::TinyVector<int,2> j = i - 1 + b();
                    stencil(b()) = phi(j);
                }

                const double* rhs = stencil.data();
                for (int i = 0; i < 16; ++i)
                {
                    c(i) = 0.0;
                    for (int j = 0; j < 16; ++j)
                        c(i) += bicubicInverse[i*16+j]*rhs[j];
                }

                blitz::TinyVector<double,2> dxinv = 1.0 / dx;
                double sx = 1.0;
                for (int i = 0; i < 4; ++i)
                {
                    double sy = 1.0;
                    for (int j = 0; j < 4; ++j)
                    {
                        c(i*4+j) *= sx*sy;
                        sy *= dxinv(1);
                    }
                    sx *= dxinv(0);
                }
            }
        
            double operator() (const blitz::TinyVector<double,2>& x) const
            {
                return c(0) + x(1)*(c(1) + x(1)*(c(2) + x(1)*c(3))) + x(0)*(c(4) + x(1)*(c(5) + x(1)*(c(6) + x(1)*c(7))) + x(0)*(c(8) + x(1)*(c(9) + x(1)*(c(10) + x(1)*c(11))) + x(0)*(c(12) + x(1)*(c(13) + x(1)*(c(14) + x(1)*c(15))))));
            }

            blitz::TinyVector<double,2> grad(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,2>(c(4) + x(1)*(c(5) + x(1)*(c(6) + x(1)*c(7))) + x(0)*(2.0*c(8) + x(1)*(2.0*c(9) + x(1)*(2.0*c(10) + x(1)*(2.0*c(11)))) + x(0)*(3.0*c(12) + x(1)*(3.0*c(13) + x(1)*(3.0*c(14) + x(1)*(3.0*c(15)))))),
                    c(1) + x(1)*(2.0*c(2) + x(1)*(3.0*c(3))) + x(0)*(c(5) + x(1)*(2.0*c(6) + x(1)*(3.0*c(7))) + x(0)*(c(9) + x(1)*(2.0*c(10) + x(1)*(3.0*c(11))) + x(0)*(c(13) + x(1)*(2.0*c(14) + x(1)*(3.0*c(15)))))));
            }

            blitz::TinyVector<double,3> hessian(const blitz::TinyVector<double,2>& x) const
            {
                return blitz::TinyVector<double,3>(2.0*c(8) + x(1)*(2.0*c(9) + x(1)*(2.0*c(10) + x(1)*(2.0*c(11)))) + x(0)*(6.0*c(12) + x(1)*(6.0*c(13) + x(1)*(6.0*c(14) + x(1)*(6.0*c(15))))),
                    c(5) + x(1)*(2.0*c(6) + x(1)*(3.0*c(7))) + x(0)*(2.0*c(9) + x(1)*(4.0*c(10) + x(1)*(6.0*c(11))) + x(0)*(3.0*c(13) + x(1)*(6.0*c(14) + x(1)*(9.0*c(15))))),
                    2.0*c(2) + x(1)*(6.0*c(3)) + x(0)*(2.0*c(6) + x(1)*(6.0*c(7)) + x(0)*(2.0*c(10) + x(1)*(6.0*c(11)) + x(0)*(2.0*c(14) + x(1)*(6.0*c(15))))));
            }
        };

    } // namespace Poly2D

    // Three-dimensional polynomials have their own namespace
    namespace Poly3D
    {
        // The data for these are defined in poly.cpp
        extern const blitz::TinyVector<int,3> stencil32points[32];
        extern const blitz::TinyVector<int,3> stencil88points[88];
        extern const double polyDegree2Inverse[10*32];
        extern const double polyDegree3Inverse[20*32];
        extern const double polyDegree4Inverse[35*88];
        extern const double polyDegree5Inverse[56*88];
        extern const double tricubicInverse[64*64];

        /* Three-dimensional, degree 2 polynomial of the form
             p(x,y) = c0 + c1 x + c2 y + c3 z + c4 x^2 + c5 x y + c6 x z + c7 y^2 + c8 y z + c9 z^2
           based on a 32-point stencil. Pseudoinverse has rank 10. */
        struct PolyDegree2
        {
            enum { order = 3 };

            blitz::TinyVector<double,10> c;

            PolyDegree2() {}

            template<typename F>
            PolyDegree2(const blitz::TinyVector<int,3>& i, const F& phi, const blitz::TinyVector<double,3>& dx)
            {
                calculateCoefficients<3,10,32>(c, stencil32points, polyDegree2Inverse, i, phi);
                blitz::TinyVector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
            }

            double operator() (const blitz::TinyVector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*c(9)) + x(1)*(c(2) + x(2)*c(8) + x(1)*c(7)) + x(0)*(c(1) + x(2)*c(6) + x(1)*c(5) + x(0)*c(4));
            }

            blitz::TinyVector<double,3> grad(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,3>(c(1) + x(2)*c(6) + x(1)*c(5) + x(0)*(2.0*c(4)), c(2) + x(2)*c(8) + x(1)*(2.0*c(7)) + x(0)*c(5), c(3) + x(2)*(2.0*c(9)) + x(1)*c(8) + x(0)*c(6));
            }

            blitz::TinyVector<double,6> hessian(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,6>(2.0*c(4), c(5), c(6), 2.0*c(7), c(8), 2.0*c(9));
            }
        };

        /* Three-dimensional, degree 3 polynomial of the form
             p(x,y) = c0 + c1 x + c2 y + c3 z + c4 x^2 + c5 x y + c6 x z + c7 y^2 + c8 y z + c9 z^2 +
                      c10 x^3 + c11 x^2 y + c12 x^2 z + c13 x y^2 + c14 x y z + c15 x z^2 + c16 y^3 + c17 y^2 z + c18 y z^2 + c19 z^3
           based on a 32-point stencil. Pseudoinverse has rank 20. */
        struct PolyDegree3
        {
            enum { order = 4 };

            blitz::TinyVector<double,20> c;

            PolyDegree3() {}

            template<typename F>
            PolyDegree3(const blitz::TinyVector<int,3>& i, const F& phi, const blitz::TinyVector<double,3>& dx)
            {
                calculateCoefficients<3,20,32>(c, stencil32points, polyDegree3Inverse, i, phi);
                blitz::TinyVector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(2);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(0)*dxinv(1)*dxinv(2);
                c(15) *= dxinv(0)*dxinv(2)*dxinv(2);
                c(16) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(17) *= dxinv(1)*dxinv(1)*dxinv(2);
                c(18) *= dxinv(1)*dxinv(2)*dxinv(2);
                c(19) *= dxinv(2)*dxinv(2)*dxinv(2);
            }

            double operator() (const blitz::TinyVector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*(c(9) + x(2)*c(19))) + x(1)*(c(2) + x(2)*(c(8) + x(2)*c(18)) + x(1)*(c(7) + x(2)*c(17) + x(1)*c(16))) + x(0)*(c(1) + x(2)*(c(6) + x(2)*c(15)) + x(1)*(c(5) + x(2)*c(14) + x(1)*c(13)) + x(0)*(c(4) + x(2)*c(12) + x(1)*c(11) + x(0)*c(10)));
            }

            blitz::TinyVector<double,3> grad(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,3>(c(1) + x(2)*(c(6) + x(2)*c(15)) + x(1)*(c(5) + x(2)*c(14) + x(1)*c(13)) + x(0)*(2.0*c(4) + x(2)*(2.0*c(12)) + x(1)*(2.0*c(11)) + x(0)*(3.0*c(10))),
                    c(2) + x(2)*(c(8) + x(2)*c(18)) + x(1)*(2.0*c(7) + x(2)*(2.0*c(17)) + x(1)*(3.0*c(16))) + x(0)*(c(5) + x(2)*c(14) + x(1)*(2.0*c(13)) + x(0)*c(11)),
                    c(3) + x(2)*(2.0*c(9) + x(2)*(3.0*c(19))) + x(1)*(c(8) + x(2)*(2.0*c(18)) + x(1)*c(17)) + x(0)*(c(6) + x(2)*(2.0*c(15)) + x(1)*c(14) + x(0)*c(12)));
            }

            blitz::TinyVector<double,6> hessian(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,6>(2.0*c(4) + x(2)*(2.0*c(12)) + x(1)*(2.0*c(11)) + x(0)*(6.0*c(10)), c(5) + x(2)*c(14) + x(1)*(2.0*c(13)) + x(0)*(2.0*c(11)),
                    c(6) + x(2)*(2.0*c(15)) + x(1)*c(14) + x(0)*(2.0*c(12)), 2.0*c(7) + x(2)*(2.0*c(17)) + x(1)*(6.0*c(16)) + x(0)*(2.0*c(13)),
                    c(8) + x(2)*(2.0*c(18)) + x(1)*(2.0*c(17)) + x(0)*c(14), 2.0*c(9) + x(2)*(6.0*c(19)) + x(1)*(2.0*c(18)) + x(0)*(2.0*c(15)));
            }
        };

        /* Three-dimensional, degree 4 polynomial of the form
             p(x,y) = c0 + c1 x + c2 y + c3 z + c4 x^2 + c5 x y + c6 x z + c7 y^2 + c8 y z + c9 z^2 +
                      c10 x^3 + c11 x^2 y + c12 x^2 z + c13 x y^2 + c14 x y z + c15 x z^2 + c16 y^3 + c17 y^2 z + c18 y z^2 + c19 z^3 +
                      c20 x^4 + c21 x^3 y + c22 x^3 z + c23 x^2 y^2 + c24 x^2 y z + c25 x^2 z^2 + c26 x y^3 + c27 x y^2 z + c28 x y z^2 +
                      c29 x z^3 + c30 y^4 + c31 y^3 z + c32 y^2 z^2 + c33 y z^3 + c34 z^4
           based on an 88-point stencil. Pseudoinverse has rank 35. */
        struct PolyDegree4
        {
            enum { order = 5 };

            blitz::TinyVector<double,35> c;

            PolyDegree4() {}

            template<typename F>
            PolyDegree4(const blitz::TinyVector<int,3>& i, const F& phi, const blitz::TinyVector<double,3>& dx)
            {
                calculateCoefficients<3,35,88>(c, stencil88points, polyDegree4Inverse, i, phi);
                blitz::TinyVector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(2);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(0)*dxinv(1)*dxinv(2);
                c(15) *= dxinv(0)*dxinv(2)*dxinv(2);
                c(16) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(17) *= dxinv(1)*dxinv(1)*dxinv(2);
                c(18) *= dxinv(1)*dxinv(2)*dxinv(2);
                c(19) *= dxinv(2)*dxinv(2)*dxinv(2);
                c(20) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(21) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(22) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2);
                c(23) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(24) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2);
                c(25) *= dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2);
                c(26) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(27) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2);
                c(28) *= dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2);
                c(29) *= dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2);
                c(30) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(31) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(32) *= dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(33) *= dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(34) *= dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
            }

            double operator() (const blitz::TinyVector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*(c(9) + x(2)*(c(19) + x(2)*c(34)))) + x(1)*(c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*c(33))) + x(1)*(c(7) + x(2)*(c(17) + x(2)*c(32)) + x(1)*(c(16) + x(2)*c(31) + x(1)*c(30)))) + x(0)*(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*c(29))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(c(13) + x(2)*c(27) + x(1)*c(26))) + x(0)*(c(4) + x(2)*(c(12) + x(2)*c(25)) + x(1)*(c(11) + x(2)*c(24) + x(1)*c(23)) + x(0)*(c(10) + x(2)*c(22) + x(1)*c(21) + x(0)*c(20))));
            }

            blitz::TinyVector<double,3> grad(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,3>(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*c(29))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(c(13) + x(2)*c(27) + x(1)*c(26))) + x(0)*(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24)) + x(1)*(2.0*c(23))) + x(0)*(3.0*c(10) + x(2)*(3.0*c(22)) + x(1)*(3.0*c(21)) + x(0)*(4.0*c(20)))),
                    c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*c(33))) + x(1)*(2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32))) + x(1)*(3.0*c(16) + x(2)*(3.0*c(31)) + x(1)*(4.0*c(30)))) + x(0)*(c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27)) + x(1)*(3.0*c(26))) + x(0)*(c(11) + x(2)*c(24) + x(1)*(2.0*c(23)) + x(0)*c(21))),
                    c(3) + x(2)*(2.0*c(9) + x(2)*(3.0*c(19) + x(2)*(4.0*c(34)))) + x(1)*(c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33))) + x(1)*(c(17) + x(2)*(2.0*c(32)) + x(1)*c(31))) + x(0)*(c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29))) + x(1)*(c(14) + x(2)*(2.0*c(28)) + x(1)*c(27)) + x(0)*(c(12) + x(2)*(2.0*c(25)) + x(1)*c(24) + x(0)*c(22))));
            }

            blitz::TinyVector<double,6> hessian(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,6>(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24)) + x(1)*(2.0*c(23))) + x(0)*(6.0*c(10) + x(2)*(6.0*c(22)) + x(1)*(6.0*c(21)) + x(0)*(12.0*c(20))),
                    c(5) + x(2)*(c(14) + x(2)*c(28)) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27)) + x(1)*(3.0*c(26))) + x(0)*(2.0*c(11) + x(2)*(2.0*c(24)) + x(1)*(4.0*c(23)) + x(0)*(3.0*c(21))),
                    c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29))) + x(1)*(c(14) + x(2)*(2.0*c(28)) + x(1)*c(27)) + x(0)*(2.0*c(12) + x(2)*(4.0*c(25)) + x(1)*(2.0*c(24)) + x(0)*(3.0*c(22))),
                    2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32))) + x(1)*(6.0*c(16) + x(2)*(6.0*c(31)) + x(1)*(12.0*c(30))) + x(0)*(2.0*c(13) + x(2)*(2.0*c(27)) + x(1)*(6.0*c(26)) + x(0)*(2.0*c(23))),
                    c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33))) + x(1)*(2.0*c(17) + x(2)*(4.0*c(32)) + x(1)*(3.0*c(31))) + x(0)*(c(14) + x(2)*(2.0*c(28)) + x(1)*(2.0*c(27)) + x(0)*c(24)),
                    2.0*c(9) + x(2)*(6.0*c(19) + x(2)*(12.0*c(34))) + x(1)*(2.0*c(18) + x(2)*(6.0*c(33)) + x(1)*(2.0*c(32))) + x(0)*(2.0*c(15) + x(2)*(6.0*c(29)) + x(1)*(2.0*c(28)) + x(0)*(2.0*c(25))));
            }
        };

        /* Three-dimensional, degree 5 polynomial of the form
             p(x,y) = c0 + c1 x + c2 y + c3 z + ... + c55 z^5
           based on an 88-point stencil. Pseudoinverse has rank 56. */
        struct PolyDegree5
        {
            enum { order = 6 };

            blitz::TinyVector<double,56> c;

            PolyDegree5() {}

            template<typename F>
            PolyDegree5(const blitz::TinyVector<int,3>& i, const F& phi, const blitz::TinyVector<double,3>& dx)
            {
                calculateCoefficients<3,56,88>(c, stencil88points, polyDegree5Inverse, i, phi);
                blitz::TinyVector<double,3> dxinv = 1.0 / dx;
                c(1) *= dxinv(0);
                c(2) *= dxinv(1);
                c(3) *= dxinv(2);
                c(4) *= dxinv(0)*dxinv(0);
                c(5) *= dxinv(0)*dxinv(1);
                c(6) *= dxinv(0)*dxinv(2);
                c(7) *= dxinv(1)*dxinv(1);
                c(8) *= dxinv(1)*dxinv(2);
                c(9) *= dxinv(2)*dxinv(2);
                c(10) *= dxinv(0)*dxinv(0)*dxinv(0);
                c(11) *= dxinv(0)*dxinv(0)*dxinv(1);
                c(12) *= dxinv(0)*dxinv(0)*dxinv(2);
                c(13) *= dxinv(0)*dxinv(1)*dxinv(1);
                c(14) *= dxinv(0)*dxinv(1)*dxinv(2);
                c(15) *= dxinv(0)*dxinv(2)*dxinv(2);
                c(16) *= dxinv(1)*dxinv(1)*dxinv(1);
                c(17) *= dxinv(1)*dxinv(1)*dxinv(2);
                c(18) *= dxinv(1)*dxinv(2)*dxinv(2);
                c(19) *= dxinv(2)*dxinv(2)*dxinv(2);
                c(20) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(21) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(22) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2);
                c(23) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(24) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2);
                c(25) *= dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2);
                c(26) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(27) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2);
                c(28) *= dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2);
                c(29) *= dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2);
                c(30) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(31) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(32) *= dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(33) *= dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(34) *= dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
                c(35) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0);
                c(36) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1);
                c(37) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2);
                c(38) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1);
                c(39) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2);
                c(40) *= dxinv(0)*dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2);
                c(41) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1);
                c(42) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2);
                c(43) *= dxinv(0)*dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2);
                c(44) *= dxinv(0)*dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2);
                c(45) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(46) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(47) *= dxinv(0)*dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(48) *= dxinv(0)*dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(49) *= dxinv(0)*dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
                c(50) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1);
                c(51) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2);
                c(52) *= dxinv(1)*dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2);
                c(53) *= dxinv(1)*dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2);
                c(54) *= dxinv(1)*dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
                c(55) *= dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2)*dxinv(2);
            }

            double operator() (const blitz::TinyVector<double,3>& x) const
            {
                return c(0) + x(2)*(c(3) + x(2)*(c(9) + x(2)*(c(19) + x(2)*(c(34) + x(2)*c(55))))) + x(1)*(c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*(c(33) + x(2)*c(54)))) + x(1)*(c(7) + x(2)*(c(17) + x(2)*(c(32) + x(2)*c(53))) + x(1)*(c(16) + x(2)*(c(31) + x(2)*c(52)) + x(1)*(c(30) + x(2)*c(51) + x(1)*c(50))))) + x(0)*(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*(c(29) + x(2)*c(49)))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(c(13) + x(2)*(c(27) + x(2)*c(47)) + x(1)*(c(26) + x(2)*c(46) + x(1)*c(45)))) + x(0)*(c(4) + x(2)*(c(12) + x(2)*(c(25) + x(2)*c(44))) + x(1)*(c(11) + x(2)*(c(24) + x(2)*c(43)) + x(1)*(c(23) + x(2)*c(42) + x(1)*c(41))) + x(0)*(c(10) + x(2)*(c(22) + x(2)*c(40)) + x(1)*(c(21) + x(2)*c(39) + x(1)*c(38)) + x(0)*(c(20) + x(2)*c(37) + x(1)*c(36) + x(0)*c(35)))));
            }

            blitz::TinyVector<double,3> grad(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,3>(c(1) + x(2)*(c(6) + x(2)*(c(15) + x(2)*(c(29) + x(2)*c(49)))) + x(1)*(c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(c(13) + x(2)*(c(27) + x(2)*c(47)) + x(1)*(c(26) + x(2)*c(46) + x(1)*c(45)))) + x(0)*(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25) + x(2)*(2.0*c(44)))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24) + x(2)*(2.0*c(43))) + x(1)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(2.0*c(41)))) + x(0)*(3.0*c(10) + x(2)*(3.0*c(22) + x(2)*(3.0*c(40))) + x(1)*(3.0*c(21) + x(2)*(3.0*c(39)) + x(1)*(3.0*c(38))) + x(0)*(4.0*c(20) + x(2)*(4.0*c(37)) + x(1)*(4.0*c(36)) + x(0)*(5.0*c(35))))),
                    c(2) + x(2)*(c(8) + x(2)*(c(18) + x(2)*(c(33) + x(2)*c(54)))) + x(1)*(2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32) + x(2)*(2.0*c(53)))) + x(1)*(3.0*c(16) + x(2)*(3.0*c(31) + x(2)*(3.0*c(52))) + x(1)*(4.0*c(30) + x(2)*(4.0*c(51)) + x(1)*(5.0*c(50))))) + x(0)*(c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27) + x(2)*(2.0*c(47))) + x(1)*(3.0*c(26) + x(2)*(3.0*c(46)) + x(1)*(4.0*c(45)))) + x(0)*(c(11) + x(2)*(c(24) + x(2)*c(43)) + x(1)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(3.0*c(41))) + x(0)*(c(21) + x(2)*c(39) + x(1)*(2.0*c(38)) + x(0)*c(36)))),
                    c(3) + x(2)*(2.0*c(9) + x(2)*(3.0*c(19) + x(2)*(4.0*c(34) + x(2)*(5.0*c(55))))) + x(1)*(c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33) + x(2)*(4.0*c(54)))) + x(1)*(c(17) + x(2)*(2.0*c(32) + x(2)*(3.0*c(53))) + x(1)*(c(31) + x(2)*(2.0*c(52)) + x(1)*c(51)))) + x(0)*(c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29) + x(2)*(4.0*c(49)))) + x(1)*(c(14) + x(2)*(2.0*c(28) + x(2)*(3.0*c(48))) + x(1)*(c(27) + x(2)*(2.0*c(47)) + x(1)*c(46))) + x(0)*(c(12) + x(2)*(2.0*c(25) + x(2)*(3.0*c(44))) + x(1)*(c(24) + x(2)*(2.0*c(43)) + x(1)*c(42)) + x(0)*(c(22) + x(2)*(2.0*c(40)) + x(1)*c(39) + x(0)*c(37)))));
            }

            blitz::TinyVector<double,6> hessian(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,6>(2.0*c(4) + x(2)*(2.0*c(12) + x(2)*(2.0*c(25) + x(2)*(2.0*c(44)))) + x(1)*(2.0*c(11) + x(2)*(2.0*c(24) + x(2)*(2.0*c(43))) + x(1)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(2.0*c(41)))) + x(0)*(6.0*c(10) + x(2)*(6.0*c(22) + x(2)*(6.0*c(40))) + x(1)*(6.0*c(21) + x(2)*(6.0*c(39)) + x(1)*(6.0*c(38))) + x(0)*(12.0*c(20) + x(2)*(12.0*c(37)) + x(1)*(12.0*c(36)) + x(0)*(20.0*c(35)))),
                    c(5) + x(2)*(c(14) + x(2)*(c(28) + x(2)*c(48))) + x(1)*(2.0*c(13) + x(2)*(2.0*c(27) + x(2)*(2.0*c(47))) + x(1)*(3.0*c(26) + x(2)*(3.0*c(46)) + x(1)*(4.0*c(45)))) + x(0)*(2.0*c(11) + x(2)*(2.0*c(24) + x(2)*(2.0*c(43))) + x(1)*(4.0*c(23) + x(2)*(4.0*c(42)) + x(1)*(6.0*c(41))) + x(0)*(3.0*c(21) + x(2)*(3.0*c(39)) + x(1)*(6.0*c(38)) + x(0)*(4.0*c(36)))),
                    c(6) + x(2)*(2.0*c(15) + x(2)*(3.0*c(29) + x(2)*(4.0*c(49)))) + x(1)*(c(14) + x(2)*(2.0*c(28) + x(2)*(3.0*c(48))) + x(1)*(c(27) + x(2)*(2.0*c(47)) + x(1)*c(46))) + x(0)*(2.0*c(12) + x(2)*(4.0*c(25) + x(2)*(6.0*c(44))) + x(1)*(2.0*c(24) + x(2)*(4.0*c(43)) + x(1)*(2.0*c(42))) + x(0)*(3.0*c(22) + x(2)*(6.0*c(40)) + x(1)*(3.0*c(39)) + x(0)*(4.0*c(37)))),
                    2.0*c(7) + x(2)*(2.0*c(17) + x(2)*(2.0*c(32) + x(2)*(2.0*c(53)))) + x(1)*(6.0*c(16) + x(2)*(6.0*c(31) + x(2)*(6.0*c(52))) + x(1)*(12.0*c(30) + x(2)*(12.0*c(51)) + x(1)*(20.0*c(50)))) + x(0)*(2.0*c(13) + x(2)*(2.0*c(27) + x(2)*(2.0*c(47))) + x(1)*(6.0*c(26) + x(2)*(6.0*c(46)) + x(1)*(12.0*c(45))) + x(0)*(2.0*c(23) + x(2)*(2.0*c(42)) + x(1)*(6.0*c(41)) + x(0)*(2.0*c(38)))),
                    c(8) + x(2)*(2.0*c(18) + x(2)*(3.0*c(33) + x(2)*(4.0*c(54)))) + x(1)*(2.0*c(17) + x(2)*(4.0*c(32) + x(2)*(6.0*c(53))) + x(1)*(3.0*c(31) + x(2)*(6.0*c(52)) + x(1)*(4.0*c(51)))) + x(0)*(c(14) + x(2)*(2.0*c(28) + x(2)*(3.0*c(48))) + x(1)*(2.0*c(27) + x(2)*(4.0*c(47)) + x(1)*(3.0*c(46))) + x(0)*(c(24) + x(2)*(2.0*c(43)) + x(1)*(2.0*c(42)) + x(0)*c(39))),
                    2.0*c(9) + x(2)*(6.0*c(19) + x(2)*(12.0*c(34) + x(2)*(20.0*c(55)))) + x(1)*(2.0*c(18) + x(2)*(6.0*c(33) + x(2)*(12.0*c(54))) + x(1)*(2.0*c(32) + x(2)*(6.0*c(53)) + x(1)*(2.0*c(52)))) + x(0)*(2.0*c(15) + x(2)*(6.0*c(29) + x(2)*(12.0*c(49))) + x(1)*(2.0*c(28) + x(2)*(6.0*c(48)) + x(1)*(2.0*c(47))) + x(0)*(2.0*c(25) + x(2)*(6.0*c(44)) + x(1)*(2.0*c(43)) + x(0)*(2.0*c(40)))));
            }
        };

        /* Three-dimensional, tricubic interpolant guaranteeing C^1 continuity between cells on a rectangular
           grid, using second order finite differences evaluated at gird points to ensure continuity of
           the gradient across grid cells. This interpolant is the analogue of the bicubic in two dimensions. */
        struct Tricubic
        {
            enum { order = 3 };

            blitz::TinyVector<double,64> c;

            template<typename F>
            Tricubic(const blitz::TinyVector<int,3>& i, const F& phi, const blitz::TinyVector<double,3>& dx)
            {
                blitz::TinyVector<int,3> stencilext = 4;
                blitz::Array<double,3> stencil(stencilext);

                for (MultiLoop<3> b(0, 4); b; ++b)
                {
                    blitz::TinyVector<int,3> j = i - 1 + b();
                    stencil(b()) = phi(j);
                }

                const double* rhs = stencil.data();
                for (int i = 0; i < 64; ++i)
                {
                    c(i) = 0.0;
                    for (int j = 0; j < 64; ++j)
                        c(i) += tricubicInverse[i*64+j]*rhs[j];
                }

                blitz::TinyVector<double,3> dxinv = 1.0 / dx;
                double sx = 1.0;
                for (int i = 0; i < 4; ++i)
                {
                    double sy = 1.0;
                    for (int j = 0; j < 4; ++j)
                    {
                        double sz = 1.0;
                        for (int k = 0; k < 4; ++k)
                        {
                            c(i*16+j*4+k) *= sx*sy*sz;
                            sz *= dxinv(2);
                        }
                        sy *= dxinv(1);
                    }
                    sx *= dxinv(0);
                }
            }

            double operator() (const blitz::TinyVector<double,3>& x) const
            {
                return c(0) + x(2)*(c(1) + x(2)*(c(2) + x(2)*c(3))) + x(1)*(c(4) + x(2)*(c(5) + x(2)*(c(6) + x(2)*c(7))) + x(1)*(c(8) + x(2)*(c(9) + x(2)*(c(10) + x(2)*c(11))) + x(1)*(c(12) + x(2)*(c(13) + x(2)*(c(14) + x(2)*c(15)))))) + x(0)*(c(16) + x(2)*(c(17) + x(2)*(c(18) + x(2)*c(19))) + x(1)*(c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(c(24) + x(2)*(c(25) + x(2)*(c(26) + x(2)*c(27))) + x(1)*(c(28) + x(2)*(c(29) + x(2)*(c(30) + x(2)*c(31)))))) + x(0)*(c(32) + x(2)*(c(33) + x(2)*(c(34) + x(2)*c(35))) + x(1)*(c(36) + x(2)*(c(37) + x(2)*(c(38) + x(2)*c(39))) + x(1)*(c(40) + x(2)*(c(41) + x(2)*(c(42) + x(2)*c(43))) + x(1)*(c(44) + x(2)*(c(45) + x(2)*(c(46) + x(2)*c(47)))))) + x(0)*(c(48) + x(2)*(c(49) + x(2)*(c(50) + x(2)*c(51))) + x(1)*(c(52) + x(2)*(c(53) + x(2)*(c(54) + x(2)*c(55))) + x(1)*(c(56) + x(2)*(c(57) + x(2)*(c(58) + x(2)*c(59))) + x(1)*(c(60) + x(2)*(c(61) + x(2)*(c(62) + x(2)*c(63)))))))));
            }

            blitz::TinyVector<double,3> grad(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,3>(c(16) + x(2)*(c(17) + x(2)*(c(18) + x(2)*c(19))) + x(1)*(c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(c(24) + x(2)*(c(25) + x(2)*(c(26) + x(2)*c(27))) + x(1)*(c(28) + x(2)*(c(29) + x(2)*(c(30) + x(2)*c(31)))))) + x(0)*(2.0*c(32) + x(2)*(2.0*c(33) + x(2)*(2.0*c(34) + x(2)*(2.0*c(35)))) + x(1)*(2.0*c(36) + x(2)*(2.0*c(37) + x(2)*(2.0*c(38) + x(2)*(2.0*c(39)))) + x(1)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(2.0*c(44) + x(2)*(2.0*c(45) + x(2)*(2.0*c(46) + x(2)*(2.0*c(47))))))) + x(0)*(3.0*c(48) + x(2)*(3.0*c(49) + x(2)*(3.0*c(50) + x(2)*(3.0*c(51)))) + x(1)*(3.0*c(52) + x(2)*(3.0*c(53) + x(2)*(3.0*c(54) + x(2)*(3.0*c(55)))) + x(1)*(3.0*c(56) + x(2)*(3.0*c(57) + x(2)*(3.0*c(58) + x(2)*(3.0*c(59)))) + x(1)*(3.0*c(60) + x(2)*(3.0*c(61) + x(2)*(3.0*c(62) + x(2)*(3.0*c(63))))))))),
                    c(4) + x(2)*(c(5) + x(2)*(c(6) + x(2)*c(7))) + x(1)*(2.0*c(8) + x(2)*(2.0*c(9) + x(2)*(2.0*c(10) + x(2)*(2.0*c(11)))) + x(1)*(3.0*c(12) + x(2)*(3.0*c(13) + x(2)*(3.0*c(14) + x(2)*(3.0*c(15)))))) + x(0)*(c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(2.0*c(24) + x(2)*(2.0*c(25) + x(2)*(2.0*c(26) + x(2)*(2.0*c(27)))) + x(1)*(3.0*c(28) + x(2)*(3.0*c(29) + x(2)*(3.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(c(36) + x(2)*(c(37) + x(2)*(c(38) + x(2)*c(39))) + x(1)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(3.0*c(44) + x(2)*(3.0*c(45) + x(2)*(3.0*c(46) + x(2)*(3.0*c(47)))))) + x(0)*(c(52) + x(2)*(c(53) + x(2)*(c(54) + x(2)*c(55))) + x(1)*(2.0*c(56) + x(2)*(2.0*c(57) + x(2)*(2.0*c(58) + x(2)*(2.0*c(59)))) + x(1)*(3.0*c(60) + x(2)*(3.0*c(61) + x(2)*(3.0*c(62) + x(2)*(3.0*c(63))))))))),
                    c(1) + x(2)*(2.0*c(2) + x(2)*(3.0*c(3))) + x(1)*(c(5) + x(2)*(2.0*c(6) + x(2)*(3.0*c(7))) + x(1)*(c(9) + x(2)*(2.0*c(10) + x(2)*(3.0*c(11))) + x(1)*(c(13) + x(2)*(2.0*c(14) + x(2)*(3.0*c(15)))))) + x(0)*(c(17) + x(2)*(2.0*c(18) + x(2)*(3.0*c(19))) + x(1)*(c(21) + x(2)*(2.0*c(22) + x(2)*(3.0*c(23))) + x(1)*(c(25) + x(2)*(2.0*c(26) + x(2)*(3.0*c(27))) + x(1)*(c(29) + x(2)*(2.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(c(33) + x(2)*(2.0*c(34) + x(2)*(3.0*c(35))) + x(1)*(c(37) + x(2)*(2.0*c(38) + x(2)*(3.0*c(39))) + x(1)*(c(41) + x(2)*(2.0*c(42) + x(2)*(3.0*c(43))) + x(1)*(c(45) + x(2)*(2.0*c(46) + x(2)*(3.0*c(47)))))) + x(0)*(c(49) + x(2)*(2.0*c(50) + x(2)*(3.0*c(51))) + x(1)*(c(53) + x(2)*(2.0*c(54) + x(2)*(3.0*c(55))) + x(1)*(c(57) + x(2)*(2.0*c(58) + x(2)*(3.0*c(59))) + x(1)*(c(61) + x(2)*(2.0*c(62) + x(2)*(3.0*c(63))))))))));
            }

            blitz::TinyVector<double,6> hessian(const blitz::TinyVector<double,3>& x) const
            {
                return blitz::TinyVector<double,6>(2.0*c(32) + x(2)*(2.0*c(33) + x(2)*(2.0*c(34) + x(2)*(2.0*c(35)))) + x(1)*(2.0*c(36) + x(2)*(2.0*c(37) + x(2)*(2.0*c(38) + x(2)*(2.0*c(39)))) + x(1)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(2.0*c(44) + x(2)*(2.0*c(45) + x(2)*(2.0*c(46) + x(2)*(2.0*c(47))))))) + x(0)*(6.0*c(48) + x(2)*(6.0*c(49) + x(2)*(6.0*c(50) + x(2)*(6.0*c(51)))) + x(1)*(6.0*c(52) + x(2)*(6.0*c(53) + x(2)*(6.0*c(54) + x(2)*(6.0*c(55)))) + x(1)*(6.0*c(56) + x(2)*(6.0*c(57) + x(2)*(6.0*c(58) + x(2)*(6.0*c(59)))) + x(1)*(6.0*c(60) + x(2)*(6.0*c(61) + x(2)*(6.0*c(62) + x(2)*(6.0*c(63)))))))),
                    c(20) + x(2)*(c(21) + x(2)*(c(22) + x(2)*c(23))) + x(1)*(2.0*c(24) + x(2)*(2.0*c(25) + x(2)*(2.0*c(26) + x(2)*(2.0*c(27)))) + x(1)*(3.0*c(28) + x(2)*(3.0*c(29) + x(2)*(3.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(2.0*c(36) + x(2)*(2.0*c(37) + x(2)*(2.0*c(38) + x(2)*(2.0*c(39)))) + x(1)*(4.0*c(40) + x(2)*(4.0*c(41) + x(2)*(4.0*c(42) + x(2)*(4.0*c(43)))) + x(1)*(6.0*c(44) + x(2)*(6.0*c(45) + x(2)*(6.0*c(46) + x(2)*(6.0*c(47)))))) + x(0)*(3.0*c(52) + x(2)*(3.0*c(53) + x(2)*(3.0*c(54) + x(2)*(3.0*c(55)))) + x(1)*(6.0*c(56) + x(2)*(6.0*c(57) + x(2)*(6.0*c(58) + x(2)*(6.0*c(59)))) + x(1)*(9.0*c(60) + x(2)*(9.0*c(61) + x(2)*(9.0*c(62) + x(2)*(9.0*c(63)))))))),
                    c(17) + x(2)*(2.0*c(18) + x(2)*(3.0*c(19))) + x(1)*(c(21) + x(2)*(2.0*c(22) + x(2)*(3.0*c(23))) + x(1)*(c(25) + x(2)*(2.0*c(26) + x(2)*(3.0*c(27))) + x(1)*(c(29) + x(2)*(2.0*c(30) + x(2)*(3.0*c(31)))))) + x(0)*(2.0*c(33) + x(2)*(4.0*c(34) + x(2)*(6.0*c(35))) + x(1)*(2.0*c(37) + x(2)*(4.0*c(38) + x(2)*(6.0*c(39))) + x(1)*(2.0*c(41) + x(2)*(4.0*c(42) + x(2)*(6.0*c(43))) + x(1)*(2.0*c(45) + x(2)*(4.0*c(46) + x(2)*(6.0*c(47)))))) + x(0)*(3.0*c(49) + x(2)*(6.0*c(50) + x(2)*(9.0*c(51))) + x(1)*(3.0*c(53) + x(2)*(6.0*c(54) + x(2)*(9.0*c(55))) + x(1)*(3.0*c(57) + x(2)*(6.0*c(58) + x(2)*(9.0*c(59))) + x(1)*(3.0*c(61) + x(2)*(6.0*c(62) + x(2)*(9.0*c(63)))))))),
                    2.0*c(8) + x(2)*(2.0*c(9) + x(2)*(2.0*c(10) + x(2)*(2.0*c(11)))) + x(1)*(6.0*c(12) + x(2)*(6.0*c(13) + x(2)*(6.0*c(14) + x(2)*(6.0*c(15))))) + x(0)*(2.0*c(24) + x(2)*(2.0*c(25) + x(2)*(2.0*c(26) + x(2)*(2.0*c(27)))) + x(1)*(6.0*c(28) + x(2)*(6.0*c(29) + x(2)*(6.0*c(30) + x(2)*(6.0*c(31))))) + x(0)*(2.0*c(40) + x(2)*(2.0*c(41) + x(2)*(2.0*c(42) + x(2)*(2.0*c(43)))) + x(1)*(6.0*c(44) + x(2)*(6.0*c(45) + x(2)*(6.0*c(46) + x(2)*(6.0*c(47))))) + x(0)*(2.0*c(56) + x(2)*(2.0*c(57) + x(2)*(2.0*c(58) + x(2)*(2.0*c(59)))) + x(1)*(6.0*c(60) + x(2)*(6.0*c(61) + x(2)*(6.0*c(62) + x(2)*(6.0*c(63)))))))),
                    c(5) + x(2)*(2.0*c(6) + x(2)*(3.0*c(7))) + x(1)*(2.0*c(9) + x(2)*(4.0*c(10) + x(2)*(6.0*c(11))) + x(1)*(3.0*c(13) + x(2)*(6.0*c(14) + x(2)*(9.0*c(15))))) + x(0)*(c(21) + x(2)*(2.0*c(22) + x(2)*(3.0*c(23))) + x(1)*(2.0*c(25) + x(2)*(4.0*c(26) + x(2)*(6.0*c(27))) + x(1)*(3.0*c(29) + x(2)*(6.0*c(30) + x(2)*(9.0*c(31))))) + x(0)*(c(37) + x(2)*(2.0*c(38) + x(2)*(3.0*c(39))) + x(1)*(2.0*c(41) + x(2)*(4.0*c(42) + x(2)*(6.0*c(43))) + x(1)*(3.0*c(45) + x(2)*(6.0*c(46) + x(2)*(9.0*c(47))))) + x(0)*(c(53) + x(2)*(2.0*c(54) + x(2)*(3.0*c(55))) + x(1)*(2.0*c(57) + x(2)*(4.0*c(58) + x(2)*(6.0*c(59))) + x(1)*(3.0*c(61) + x(2)*(6.0*c(62) + x(2)*(9.0*c(63)))))))),
                    2.0*c(2) + x(2)*(6.0*c(3)) + x(1)*(2.0*c(6) + x(2)*(6.0*c(7)) + x(1)*(2.0*c(10) + x(2)*(6.0*c(11)) + x(1)*(2.0*c(14) + x(2)*(6.0*c(15))))) + x(0)*(2.0*c(18) + x(2)*(6.0*c(19)) + x(1)*(2.0*c(22) + x(2)*(6.0*c(23)) + x(1)*(2.0*c(26) + x(2)*(6.0*c(27)) + x(1)*(2.0*c(30) + x(2)*(6.0*c(31))))) + x(0)*(2.0*c(34) + x(2)*(6.0*c(35)) + x(1)*(2.0*c(38) + x(2)*(6.0*c(39)) + x(1)*(2.0*c(42) + x(2)*(6.0*c(43)) + x(1)*(2.0*c(46) + x(2)*(6.0*c(47))))) + x(0)*(2.0*c(50) + x(2)*(6.0*c(51)) + x(1)*(2.0*c(54) + x(2)*(6.0*c(55)) + x(1)*(2.0*c(58) + x(2)*(6.0*c(59)) + x(1)*(2.0*c(62) + x(2)*(6.0*c(63)))))))));
            }
        };

    } // namespace Poly3D

    /* The following templated typedefs allow a particular dimension N and particular polynomial
       specified by an integer parameter Degree to be translated to the specific polynomial
       class implemented above. Degree = -1 means use the bicubic/tricubic. If one tries to
       use an underfined polynomial, a compiler error relating to "InvalidPoly" will be generated. */

    struct InvalidPoly {};

    template<int N, int Degree>
    struct PolyType
    {
        typedef InvalidPoly T_Poly;
    };

    template<>
    struct PolyType<2,2>
    {
        typedef Poly2D::PolyDegree2 T_Poly;
    };

    template<>
    struct PolyType<2,3>
    {
        typedef Poly2D::PolyDegree3 T_Poly;
    };

    template<>
    struct PolyType<2,4>
    {
        typedef Poly2D::PolyDegree4 T_Poly;
    };

    template<>
    struct PolyType<2,5>
    {
        typedef Poly2D::PolyDegree5 T_Poly;
    };

    template<>
    struct PolyType<2,-1>
    {
        typedef Poly2D::Bicubic T_Poly;
    };

    template<>
    struct PolyType<3,2>
    {
        typedef Poly3D::PolyDegree2 T_Poly;
    };

    template<>
    struct PolyType<3,3>
    {
        typedef Poly3D::PolyDegree3 T_Poly;
    };

    template<>
    struct PolyType<3,4>
    {
        typedef Poly3D::PolyDegree4 T_Poly;
    };

    template<>
    struct PolyType<3,5>
    {
        typedef Poly3D::PolyDegree5 T_Poly;
    };

    template<>
    struct PolyType<3,-1>
    {
        typedef Poly3D::Tricubic T_Poly;
    };

} // namespace HighOrderCP

#endif
