/* This work was supported by the Director, Office of Science, Office of Advanced Scientific
   Computing Research of the U.S. Department of Energy under contract number DE-AC02-05CH11231.
   Copyright (c) 2014. */

/* testSimple.cpp
    Provides a very simple demonstration of how to reinitialise a level set function with
    the high-order closest point algorithm. */

#include "hocp.hpp"

// Two-dimensional example using a circle of radius 1
void test2d(int n)
{
    blitz::Array<double,2> phi(n, n), sdfexact(n, n);
    double dx = 4.0 / n;
    for (int i = 0; i < n; ++i)
    {
        double x = -2.0 + (i + 0.5) * dx;
        for (int j = 0; j < n; ++j)
        {
            double y = -2.0 + (j + 0.5) * dx;
            phi(i,j) = exp(x*x + y*y - 1.0) - 1.0;
            sdfexact(i,j) = sqrt(x*x + y*y) - 1.0;
        }
    }
    
    // The following differ according to what type of polynomial is used to compute the reinitialisation
    //HighOrderCP::reinit<2,-1>(phi, dx, 10.0);
    //HighOrderCP::reinit<2,2>(phi, dx, 10.0);
    HighOrderCP::reinit<2,3>(phi, dx, 10.0);
    //HighOrderCP::reinit<2,4>(phi, dx, 10.0);
    //HighOrderCP::reinit<2,5>(phi, dx, 10.0);

    double maxerr = 0.0;
    for (int i = 0; i < n; ++i)
        for (int j = 0; j < n; ++j)
            maxerr = std::max(maxerr, fabs(phi(i,j) - sdfexact(i,j)));

    std::cout << "Infinity norm of error = " << maxerr << std::endl;
}

// Three-dimensional example using a sphere of radius 1
void test3d(int n)
{
    blitz::Array<double,3> phi(n, n, n), sdfexact(n, n, n);
    double dx = 4.0 / n;
    for (int i = 0; i < n; ++i)
    {
        double x = -2.0 + (i + 0.5) * dx;
        for (int j = 0; j < n; ++j)
        {
            double y = -2.0 + (j + 0.5) * dx;
            for (int k = 0; k < n; ++k)
            {
                double z = -2.0 + (k + 0.5) * dx;
                phi(i,j,k) = exp(x*x + y*y + z*z - 1.0) - 1.0;
                sdfexact(i,j,k) = sqrt(x*x + y*y + z*z) - 1.0;
            }
        }
    }

    // The following differ according to what type of polynomial is used to compute the reinitialisation
    //HighOrderCP::reinit<3,-1>(phi, dx, 10.0);
    //HighOrderCP::reinit<3,2>(phi, dx, 10.0);
    HighOrderCP::reinit<3,3>(phi, dx, 10.0);
    //HighOrderCP::reinit<3,4>(phi, dx, 10.0);
    //HighOrderCP::reinit<3,5>(phi, dx, 10.0);

    double maxerr = 0.0;
    for (int i = 0; i < n; ++i)
        for (int j = 0; j < n; ++j)
            for (int k = 0; k < n; ++k)
                maxerr = std::max(maxerr, fabs(phi(i,j,k) - sdfexact(i,j,k)));

    std::cout << "Infinity norm of error = " << maxerr << std::endl;
}

int main(int argc, char* argv[])
{
    while (true)
    {
        int dim = 0, n = 0;
        std::cout << "Enter the dimension N = 2 or N = 3: ";
        std::cin >> dim;
        if (dim != 2 && dim != 3) break;
        std::cout << "Enter number of grid points in each dimension: ";
        std::cin >> n;
        if (n < 2) break;
        if (dim == 2)
            test2d(n);
        else
            test3d(n);
    }
    return 0;
}